const path = require('path')
const notifier = require('node-notifier')
const {charSizes} = require('./font.js')


// Ugly hack. There are two problems (not 100% sure but this is what's most likely going on)
// 1) Rollup apparently terminates the process immediately after error occurs.
//    It calls the buildEnd callback but does not give us enough time to create the notification asynchronously.
// 2) node-notifier creates the notification by calling child_process.execFile() which is not as reliable
//    as just child_process.exec(). But exec() wouldn't support multiline notifications due to inability to pass
//    EOLs in cmd arguments.
// This hack creates the notification synchronously and bypassess node-notifier's internals.
// Tested on Windows 10. Not sure if the same problem/hack occurs/is-needed on other systems too.
const os = require('os')
if (os.platform() === 'win32' && os.release().startsWith('10.')) {
	notifier.notify = function(options) {
		var {title, message, icon} = options
		var cp = require('child_process')
		var path = require('path')
		var snorePath = './node_modules/node-notifier/vendor/snoreToast/SnoreToast.exe'
		snorePath = path.join(__dirname, snorePath)
		var args = []
		if (icon)
			args.push('-p', icon)
		if (title)
			args.push('-t', title)
		if (message)
			args.push('-m', message)
		cp.spawnSync(snorePath, args)
		return
	}
}

var icon = path.join(__dirname, 'rollup.png')

function makeDummyLine(char, size) {
	return char.repeat(Math.round(52 / size))
}

function getLength(string) {
	var characters = 0
	string
		.split('')
		.forEach(char => {
			var size = charSizes.get(char) || 1
			characters += size
		})
	return Math.round(characters)
}

function createCodeBlock(frame) {
	var lines = frame.replace(/\r/g, '').split('\n')
	var caretLineOriginal = lines.find(line => line.startsWith('  ') && line.includes('^') && !line.includes(':'))
	var codeLine = lines[lines.indexOf(caretLineOriginal) - 1]
	var caretStart = caretLineOriginal.indexOf('^')
	var caretCountOriginal = caretLineOriginal.trim().length
	var toBeReplacedBySpaces = codeLine.substr(0, caretStart)
	var spaceCount = getLength(toBeReplacedBySpaces)
	var spaces = ' '.repeat(spaceCount)
	if (caretCountOriginal === 1) {
		var carets = '^'
	} else {
		var toBeReplacedByMarkers = codeLine.substr(caretStart, caretCountOriginal)
		var caretCountNew = getLength(toBeReplacedByMarkers)
		var carets = '^'.repeat(caretCountNew / charSizes.get('^'))
	}
	var caretLine = spaces + carets
	return [codeLine, caretLine].join('\n')
}

function processError(error) {
	var {frame, codeFrame, loc, message} = error
	if (codeFrame) {
		// Babel styled codeFrame
	} else if (frame) {
		// Default code frame
		message += '\n' + createCodeBlock(frame)
	}
	var title = []
	if (loc) {
		// Default Rollup error has file path in property error.loc.file
		if (loc.file)
			title.push(sanitizePath(loc.file))
		if (loc.line)
			title.push(`line ${loc.line}`)
		if (loc.column)
			title.push(`column ${loc.column}`)
	} else {
		// Babel has file path in property error.id
		if (error.id)
			title.push(sanitizePath(error.id))
	}
	var title = title.filter(a => a).join(', ') || `Rollup error: ${error.code}`
	notifier.notify({title, message, icon})
}

function sanitizePath(path) {
	try {
		return path.parse(path).base
	} catch(err) {}
}

module.exports = function notify(options) {
	return {
		buildEnd(err) {
			if (err)
				processError(err)
		}
	}
}
